@@ -19,6 +19,7 @@ from . import (
     permas,
     ply,
     stl,
+    su2,
     svg,
     tecplot,
     tetgen,
@@ -28,12 +29,7 @@ from . import (
     wkt,
     xdmf,
 )
-from .__about__ import (
-    __original_author__,
-    __original_author_email__,
-    __version__,
-    __website__,
-)
+from .__about__ import __version__
 from ._exceptions import ReadError, WriteError
 from ._helpers import extension_to_filetype, read, write, write_points_cells
 from ._mesh import CellBlock, Mesh
@@ -58,6 +54,7 @@ __all__ = [
     "permas",
     "ply",
     "stl",
+    "su2",
     "svg",
     "tecplot",
     "tetgen",
@@ -76,7 +73,4 @@ __all__ = [
     "ReadError",
     "WriteError",
     "__version__",
-    "__original_author__",
-    "__original_author_email__",
-    "__website__",
 ]
