@@ -41,6 +41,11 @@ class Mesh:
         self.gmsh_periodic = gmsh_periodic
         self.info = info
 
+        for key, data in self.cell_data.items():
+            assert len(data) == len(cells), (
+                "Incompatible cell data. {} cell blocks, but '{}' has {} blocks."
+            ).format(len(cells), key, len(data))
+
     def __repr__(self):
         lines = [
             "<meshio mesh object>",
@@ -74,13 +79,13 @@ class Mesh:
 
         new_cells = []
         new_cell_data = {}
-        for c in self.cells:
+        for block_counter, c in enumerate(self.cells):
             if c.type not in prune_list:
                 new_cells.append(c)
-                for name, data in self.cell_data:
+                for name, data in self.cell_data.items():
                     if name not in new_cell_data:
                         new_cell_data[name] = []
-                    new_cell_data[name].append(data)
+                    new_cell_data[name] += [data[block_counter]]
 
         self.cells = new_cells
         self.cell_data = new_cell_data
@@ -121,6 +126,8 @@ class Mesh:
         write(path_or_buf, self, file_format, **kwargs)
 
     def get_cells_type(self, cell_type):
+        if not any(c.type == cell_type for c in self.cells):
+            return numpy.array([], dtype=int)
         return numpy.concatenate([c.data for c in self.cells if c.type == cell_type])
 
     def get_cell_data(self, name, cell_type):
@@ -206,10 +213,6 @@ class Mesh:
 
     def int_data_to_sets(self):
         """Convert all int data to {point,cell}_sets, where possible.
-
-        defined by int-valued cell_data, keyed (like cell_sets_dict) by name (as found
-        in field_data or constructed from the int) and cell_type.  The indices are into
-        the items of cells_dict[cell_type].
         """
         keys = []
         for key, data in self.cell_data.items():
