@@ -106,6 +106,8 @@ def read_buffer(f):
     cell_ids = []
     point_sets = {}
     cell_sets = {}
+    cell_sets_element = {}  # Handle cell sets defined in ELEMENT
+    cell_sets_element_order = []  # Order of keys is not preserved in Python 3.5
     field_data = {}
     cell_data = {}
     point_data = {}
@@ -124,31 +126,64 @@ def read_buffer(f):
         if keyword == "NODE":
             points, point_ids, line = _read_nodes(f)
         elif keyword == "ELEMENT":
-            cell_type, cells_data, ids, line = _read_cells(f, line, point_ids)
+            params_map = get_param_map(line, required_keys=["TYPE"])
+            cell_type, cells_data, ids, sets, line = _read_cells(
+                f, params_map, point_ids
+            )
             cells.append(CellBlock(cell_type, cells_data))
             cell_ids.append(ids)
+            if sets:
+                cell_sets_element.update(sets)
+                cell_sets_element_order += list(sets.keys())
         elif keyword == "NSET":
             params_map = get_param_map(line, required_keys=["NSET"])
-            set_ids, line = _read_set(f, params_map)
+            set_ids, _, line = _read_set(f, params_map)
             name = params_map["NSET"]
             point_sets[name] = numpy.array(
                 [point_ids[point_id] for point_id in set_ids], dtype="int32"
             )
         elif keyword == "ELSET":
             params_map = get_param_map(line, required_keys=["ELSET"])
-            set_ids, line = _read_set(f, params_map)
+            set_ids, set_names, line = _read_set(f, params_map)
             name = params_map["ELSET"]
             cell_sets[name] = []
-            for cell_ids_ in cell_ids:
-                cell_sets_ = numpy.array(
-                    [cell_ids_[set_id] for set_id in set_ids if set_id in cell_ids_],
-                    dtype="int32",
-                )
-                cell_sets[name].append(cell_sets_)
+            if set_ids.size:
+                for cell_ids_ in cell_ids:
+                    cell_sets_ = numpy.array(
+                        [
+                            cell_ids_[set_id]
+                            for set_id in set_ids
+                            if set_id in cell_ids_
+                        ],
+                        dtype="int32",
+                    )
+                    cell_sets[name].append(cell_sets_)
+            elif set_names:
+                for set_name in set_names:
+                    if set_name in cell_sets.keys():
+                        cell_sets[name].append(cell_sets[set_name])
+                    elif set_name in cell_sets_element.keys():
+                        cell_sets[name].append(cell_sets_element[set_name])
+                    else:
+                        raise ReadError("Unknown cell set '{}'".format(set_name))
         else:
             # There are just too many Abaqus keywords to explicitly skip them.
             line = f.readline()
 
+    # Parse cell sets defined in ELEMENT
+    for i, name in enumerate(cell_sets_element_order):
+        # Not sure whether this case would ever happen
+        if name in cell_sets.keys():
+            cell_sets[name][i] = cell_sets_element[name]
+        else:
+            cell_sets[name] = []
+            for ic in range(len(cells)):
+                cell_sets[name].append(
+                    cell_sets_element[name]
+                    if i == ic
+                    else numpy.array([], dtype="int32")
+                )
+
     return Mesh(
         points,
         cells,
@@ -180,15 +215,9 @@ def _read_nodes(f):
     return numpy.array(points, dtype=float), point_ids, line
 
 
-def _read_cells(f, line0, point_ids):
-    sline = line0.split(",")[1:]
-
-    etype_sline = sline[0].upper()
-    if "TYPE" not in etype_sline:
-        raise ReadError(etype_sline)
-
-    etype = etype_sline.split("=")[1].strip()
-    if etype not in abaqus_to_meshio_type:
+def _read_cells(f, params_map, point_ids):
+    etype = params_map["TYPE"]
+    if etype not in abaqus_to_meshio_type.keys():
         raise ReadError("Element type not available: {}".format(etype))
 
     cell_type = abaqus_to_meshio_type[etype]
@@ -210,7 +239,14 @@ def _read_cells(f, line0, point_ids):
             cells.append([point_ids[k] for k in idx[1:]])
             idx = []
             counter += 1
-    return cell_type, numpy.array(cells), cell_ids, line
+
+    cell_sets = (
+        {params_map["ELSET"]: numpy.arange(counter, dtype="int32")}
+        if "ELSET" in params_map.keys()
+        else {}
+    )
+
+    return cell_type, numpy.array(cells), cell_ids, cell_sets, line
 
 
 def get_param_map(word, required_keys=None):
@@ -254,6 +290,7 @@ def get_param_map(word, required_keys=None):
 
 def _read_set(f, params_map):
     set_ids = []
+    set_names = []
     while True:
         line = f.readline()
         if not line or line.startswith("*"):
@@ -261,22 +298,26 @@ def _read_set(f, params_map):
         if line.strip() == "":
             continue
 
-        set_ids += [int(k) for k in line.strip().strip(",").split(",")]
+        line = line.strip().strip(",").split(",")
+        if line[0].isnumeric():
+            set_ids += [int(k) for k in line]
+        else:
+            set_names.append(line[0])
 
     set_ids = numpy.array(set_ids, dtype="int32")
     if "GENERATE" in params_map:
         if len(set_ids) != 3:
             raise ReadError(set_ids)
         set_ids = numpy.arange(set_ids[0], set_ids[1] + 1, set_ids[2], dtype="int32")
-    return set_ids, line
+    return set_ids, set_names, line
 
 
-def write(filename, mesh, float_fmt=".15e", translate_cell_names=True):
+def write(filename, mesh, float_fmt=".16e", translate_cell_names=True):
     with open_file(filename, "wt") as f:
-        f.write("*Heading\n")
+        f.write("*HEADING\n")
         f.write("Abaqus DataFile Version 6.14\n")
         f.write("written by meshio v{}\n".format(__version__))
-        f.write("*Node\n")
+        f.write("*NODE\n")
         fmt = ", ".join(["{}"] + ["{:" + float_fmt + "}"] * mesh.points.shape[1]) + "\n"
         for k, x in enumerate(mesh.points):
             f.write(fmt.format(k + 1, *x))
@@ -285,33 +326,37 @@ def write(filename, mesh, float_fmt=".15e", translate_cell_names=True):
             name = (
                 meshio_to_abaqus_type[cell_type] if translate_cell_names else cell_type
             )
-            f.write("*Element,type=" + name + "\n")
+            f.write("*ELEMENT, TYPE={}\n".format(name))
             for row in node_idcs:
                 eid += 1
                 nids_strs = (str(nid + 1) for nid in row.tolist())
                 f.write(str(eid) + "," + ",".join(nids_strs) + "\n")
 
         nnl = 8
+        offset = 0
         for ic in range(len(mesh.cells)):
             for k, v in mesh.cell_sets.items():
-                els = [str(i + 1) for i in v[ic]]
-                f.write("*ELSET, ELSET=%s\n" % k)
-                f.write(
-                    ",\n".join(
-                        ",".join(els[i : i + nnl]) for i in range(0, len(els), nnl)
+                if len(v[ic]) > 0:
+                    els = [str(i + 1 + offset) for i in v[ic]]
+                    f.write("*ELSET, ELSET={}\n".format(k))
+                    f.write(
+                        ",\n".join(
+                            ",".join(els[i : i + nnl]) for i in range(0, len(els), nnl)
+                        )
+                        + "\n"
                     )
-                    + "\n"
-                )
+            offset += len(mesh.cells[ic].data)
 
         for k, v in mesh.point_sets.items():
             nds = [str(i + 1) for i in v]
-            f.write("*NSET, NSET=%s\n" % k)
+            f.write("*NSET, NSET={}\n".format(k))
             f.write(
                 ",\n".join(",".join(nds[i : i + nnl]) for i in range(0, len(nds), nnl))
                 + "\n"
             )
 
-        f.write("*end")
+        # https://github.com/nschloe/meshio/issues/747#issuecomment-643479921
+        # f.write("*END")
 
 
 register("abaqus", [".inp"], read, {"abaqus": write})
