@@ -11,6 +11,8 @@ from .._mesh import Mesh
 
 
 def read(filename):
+    import h5py
+
     f = h5py.File(filename, "r")
 
     x = f["Base"]["Zone1"]["GridCoordinates"]["CoordinateX"][" data"]
@@ -32,6 +34,8 @@ def read(filename):
 
 
 def write(filename, mesh, compression="gzip", compression_opts=4):
+    import h5py
+
     f = h5py.File(filename, "w")
 
     base = f.create_group("Base")
@@ -87,10 +91,4 @@ def write(filename, mesh, compression="gzip", compression_opts=4):
             )
 
 
-try:
-    import h5py
-# Use ModuleNotFoundError when dropping support for Python 3.5
-except ImportError:
-    pass
-else:
-    register("cgns", [".cgns"], read, {"cgns": write})
+register("cgns", [".cgns"], read, {"cgns": write})
