@@ -71,12 +71,17 @@ def read_buffer(f, is_ascii, data_size):
     cell_data = cell_data_from_raw(cells, cell_data_raw)
 
     # merge cell_tags into cell_data
-    for name, tag_dict in cell_tags.items():
-        if name not in cell_data:
-            cell_data[name] = []
-        for cell_type, _ in cells:
+    for tag_name, tag_dict in cell_tags.items():
+        if tag_name not in cell_data:
+            cell_data[tag_name] = []
+        offset = {}
+        for cell_type, cell_array in cells:
+            start = offset.setdefault(cell_type, 0)
+            end = start + len(cell_array)
+            offset[cell_type] = end
             tags = tag_dict.get(cell_type, [])
-            cell_data[name].append(tags)
+            tags = numpy.array(tags[start:end], dtype=c_int)
+            cell_data[tag_name].append(tags)
 
     return Mesh(
         points,
@@ -138,7 +143,7 @@ def _read_cells(f, cells, point_tags, is_ascii):
         line = f.readline().decode("utf-8")
 
     # restrict to the standard two data items (physical, geometrical)
-    output_cell_tags = {"gmsh:physical": {}, "gmsh:geometrical": {}}
+    output_cell_tags = {}
     for cell_type in cell_tags:
         physical = []
         geometrical = []
@@ -152,8 +157,12 @@ def _read_cells(f, cells, point_tags, is_ascii):
         physical = numpy.array(physical, dtype=c_int)
         geometrical = numpy.array(geometrical, dtype=c_int)
         if len(physical) > 0:
+            if "gmsh:physical" not in output_cell_tags:
+                output_cell_tags["gmsh:physical"] = {}
             output_cell_tags["gmsh:physical"][cell_type] = physical
         if len(geometrical) > 0:
+            if "gmsh:geometrical" not in output_cell_tags:
+                output_cell_tags["gmsh:geometrical"] = {}
             output_cell_tags["gmsh:geometrical"][cell_type] = geometrical
 
     return has_additional_tag_data, output_cell_tags
@@ -255,7 +264,7 @@ def _read_periodic(f):
     return periodic
 
 
-def write(filename, mesh, float_fmt=".15e", binary=True):
+def write(filename, mesh, float_fmt=".16e", binary=True):
     """Writes msh files, cf.
     <http://gmsh.info//doc/texinfo/gmsh.html#MSH-ASCII-file-format>.
     """
@@ -306,7 +315,7 @@ def write(filename, mesh, float_fmt=".15e", binary=True):
         _write_nodes(fh, mesh.points, float_fmt, binary)
         _write_elements(fh, cells, tag_data, binary)
         if mesh.gmsh_periodic is not None:
-            _write_periodic(fh, mesh.gmsh_periodic)
+            _write_periodic(fh, mesh.gmsh_periodic, float_fmt)
         for name, dat in mesh.point_data.items():
             _write_data(fh, "NodeData", name, dat, binary)
 
@@ -389,7 +398,7 @@ def _write_elements(fh, cells, tag_data, binary):
     fh.write(b"$EndElements\n")
 
 
-def _write_periodic(fh, periodic):
+def _write_periodic(fh, periodic, float_fmt):
     fh.write(b"$Periodic\n")
     fh.write("{}\n".format(len(periodic)).encode("utf-8"))
     for dim, (stag, mtag), affine, slave_master in periodic:
@@ -398,7 +407,7 @@ def _write_periodic(fh, periodic):
             fh.write(b"Affine ")
             affine = numpy.array(affine, dtype=float)
             affine = numpy.atleast_2d(affine.ravel())
-            numpy.savetxt(fh, affine, "%.16g")
+            numpy.savetxt(fh, affine, fmt="%" + float_fmt)
         slave_master = numpy.array(slave_master, dtype=c_int).reshape(-1, 2)
         slave_master = slave_master + 1  # Add one, Gmsh is 0-based
         fh.write("{}\n".format(len(slave_master)).encode("utf-8"))
