@@ -55,14 +55,13 @@ def read_buffer(f, is_ascii, data_size):
 
         if environ == "PhysicalNames":
             _read_physical_names(f, field_data)
-            cell_sets = {k: [] for k in field_data.keys()}
         elif environ == "Entities":
             physical_tags = _read_entities(f, is_ascii, data_size)
         elif environ == "Nodes":
             points, point_tags = _read_nodes(f, is_ascii, data_size)
         elif environ == "Elements":
-            cells, cell_tags = _read_elements(
-                f, point_tags, physical_tags, is_ascii, data_size, field_data, cell_sets
+            cells, cell_tags, cell_sets = _read_elements(
+                f, point_tags, physical_tags, is_ascii, data_size, field_data
             )
         elif environ == "Periodic":
             periodic = _read_periodic(f, is_ascii, data_size)
@@ -176,9 +175,7 @@ def _read_nodes(f, is_ascii, data_size):
     return points, tags
 
 
-def _read_elements(
-    f, point_tags, physical_tags, is_ascii, data_size, field_data, cell_sets
-):
+def _read_elements(f, point_tags, physical_tags, is_ascii, data_size, field_data):
     fromfile = partial(numpy.fromfile, sep=" " if is_ascii else "")
     c_size_t = _size_type(data_size)
 
@@ -189,22 +186,24 @@ def _read_elements(
 
     data = []
     cell_data = {}
+    cell_sets = {k: [None] * num_entity_blocks for k in field_data.keys()}
 
     for k in range(num_entity_blocks):
         # entityDim(int) entityTag(int) elementType(int) numElements(size_t)
         dim_entity, tag_entity, type_ele = fromfile(f, c_int, 3)
         (num_ele,) = fromfile(f, c_size_t, 1)
         for physical_name, cell_set in cell_sets.items():
-            if physical_tags:
-                for physical_tag in physical_tags[dim_entity][tag_entity]:
-                    cell_set.append(
-                        numpy.arange(
-                            num_ele
-                            if field_data[physical_name][0] == physical_tag
-                            else 0,
-                            dtype=type(num_ele),
-                        )
-                    )
+            cell_set[k] = numpy.arange(
+                num_ele
+                if (
+                    physical_tags
+                    and field_data[physical_name][1] == dim_entity
+                    and field_data[physical_name][0]
+                    in physical_tags[dim_entity][tag_entity]
+                )
+                else 0,
+                dtype=type(num_ele),
+            )
         tpe = _gmsh_to_meshio_type[type_ele]
         num_nodes_per_ele = num_nodes_per_cell[tpe]
         d = fromfile(f, c_size_t, int(num_ele * (1 + num_nodes_per_ele))).reshape(
@@ -243,7 +242,7 @@ def _read_elements(
             )
     cells[:] = _gmsh_to_meshio_order(cells)
 
-    return cells, cell_data
+    return cells, cell_data, cell_sets
 
 
 def _read_periodic(f, is_ascii, data_size):
@@ -274,11 +273,11 @@ def _read_periodic(f, is_ascii, data_size):
     return periodic
 
 
-def write(filename, mesh, binary=True):
-    write4_1(filename, mesh, binary=binary)
+def write(filename, mesh, float_fmt=".16e", binary=True):
+    write4_1(filename, mesh, float_fmt=float_fmt, binary=binary)
 
 
-def write4_1(filename, mesh, binary=True):
+def write4_1(filename, mesh, float_fmt=".16e", binary=True):
     """Writes msh files, cf.
     <http://gmsh.info/doc/texinfo/gmsh.html#MSH-file-format>.
     """
@@ -315,10 +314,10 @@ def write4_1(filename, mesh, binary=True):
             _write_physical_names(fh, mesh.field_data)
 
         _write_entities(fh, cells, binary)
-        _write_nodes(fh, mesh.points, mesh.cells, binary)
+        _write_nodes(fh, mesh.points, mesh.cells, float_fmt, binary)
         _write_elements(fh, cells, binary)
         if mesh.gmsh_periodic is not None:
-            _write_periodic(fh, mesh.gmsh_periodic, binary)
+            _write_periodic(fh, mesh.gmsh_periodic, float_fmt, binary)
         for name, dat in mesh.point_data.items():
             _write_data(fh, "NodeData", name, dat, binary)
 
@@ -360,7 +359,7 @@ def _write_entities(fh, cells, binary):
     return
 
 
-def _write_nodes(fh, points, cells, binary):
+def _write_nodes(fh, points, cells, float_fmt, binary):
     fh.write(b"$Nodes\n")
 
     # The entity_dim and entity_tag in the $Elements section must correspond to an
@@ -415,7 +414,7 @@ def _write_nodes(fh, points, cells, binary):
         )
         numpy.arange(1, 1 + n, dtype=c_size_t).tofile(fh, "\n", "%d")
         fh.write(b"\n")
-        numpy.savetxt(fh, points, delimiter=" ")
+        numpy.savetxt(fh, points, delimiter=" ", fmt="%" + float_fmt)
 
     fh.write(b"$EndNodes\n")
     return
@@ -504,7 +503,7 @@ def _write_elements(fh, cells, binary):
     return
 
 
-def _write_periodic(fh, periodic, binary):
+def _write_periodic(fh, periodic, float_fmt, binary):
     """write the $Periodic block
 
     specified as
@@ -527,9 +526,9 @@ def _write_periodic(fh, periodic, binary):
             ary.tofile(fh)
         else:
             ary = numpy.atleast_2d(ary)
-            fmt = "%.16g" if dtype == c_double else "%d"
-            fmt = kwargs.pop("fmt", fmt)
-            numpy.savetxt(fh, ary, fmt, **kwargs)
+            fmt = float_fmt if dtype == c_double else "d"
+            fmt = "%" + kwargs.pop("fmt", fmt)
+            numpy.savetxt(fh, ary, fmt=fmt, **kwargs)
 
     fh.write(b"$Periodic\n")
     tofile(fh, len(periodic), c_size_t)
@@ -539,7 +538,7 @@ def _write_periodic(fh, periodic, binary):
             tofile(fh, 0, c_size_t)
         else:
             tofile(fh, len(affine), c_size_t, newline=" ")
-            tofile(fh, affine, c_double, fmt="%.16g")
+            tofile(fh, affine, c_double, fmt=float_fmt)
         slave_master = numpy.array(slave_master, dtype=c_size_t)
         slave_master = slave_master.reshape(-1, 2)
         slave_master = slave_master + 1  # Add one, Gmsh is 1-based
