@@ -1,17 +1,21 @@
+import pathlib
 import struct
 
 from .._exceptions import ReadError, WriteError
 from .._helpers import register
 from . import _gmsh22, _gmsh40, _gmsh41
 
-_readers = {"2.2": _gmsh22, "4.0": _gmsh40, "4.1": _gmsh41}
+# Some mesh files out there have the version specified as version "2" when it really is
+# "2.2". Same with "4" vs "4.1".
+_readers = {"2": _gmsh22, "2.2": _gmsh22, "4.0": _gmsh40, "4": _gmsh41, "4.1": _gmsh41}
 _writers = {"2.2": _gmsh22, "4.0": _gmsh40, "4.1": _gmsh41}
 
 
 def read(filename):
     """Reads a Gmsh msh file.
     """
-    with open(filename, "rb") as f:
+    filename = pathlib.Path(filename)
+    with open(filename.as_posix(), "rb") as f:
         mesh = read_buffer(f)
     return mesh
 
@@ -83,7 +87,9 @@ def _read_header(f):
     return fmt_version, data_size, is_ascii
 
 
-def write(filename, mesh, fmt_version="4.1", binary=True):
+# Gmsh ASCII output uses `%.16g` for floating point values,
+# meshio uses same precision but exponential notation `%.16e`.
+def write(filename, mesh, fmt_version="4.1", binary=True, float_fmt=".16e"):
     """Writes a Gmsh msh file.
     """
     try:
@@ -98,7 +104,7 @@ def write(filename, mesh, fmt_version="4.1", binary=True):
                 )
             )
 
-    writer.write(filename, mesh, binary=binary)
+    writer.write(filename, mesh, binary=binary, float_fmt=float_fmt)
 
 
 register(
