@@ -190,9 +190,9 @@ def read_buffer(f):
     mesh.points_id = points_id
     mesh.cells_id = cells_id
     if len(point_refs) > 0:
-        mesh.point_data["nastran:ref"] = point_refs
+        mesh.point_data["nastran:ref"] = numpy.array(point_refs)
     if len(cell_refs) > 0:
-        mesh.cell_data["nastran:ref"] = cell_refs
+        mesh.cell_data["nastran:ref"] = [numpy.array(i) for i in cell_refs]
     return mesh
 
 
@@ -215,14 +215,14 @@ def write(filename, mesh):
         if point_refs is not None:
             for point_id, x in enumerate(points):
                 f.write(
-                    "GRID, {:d}, {:d}, {:.15e}, {:.15e}, {:.15e}\n".format(
+                    "GRID, {:d}, {:d}, {:.16e}, {:.16e}, {:.16e}\n".format(
                         point_id + 1, point_refs[point_id], x[0], x[1], x[2]
                     )
                 )
         else:
             for point_id, x in enumerate(points):
                 f.write(
-                    "GRID, {:d},, {:.15e}, {:.15e}, {:.15e}\n".format(
+                    "GRID, {:d},, {:.16e}, {:.16e}, {:.16e}\n".format(
                         point_id + 1, x[0], x[1], x[2]
                     )
                 )
