@@ -4,7 +4,6 @@ I/O for the STL format, cf.
 """
 import logging
 import os
-import re
 
 import numpy
 
@@ -45,8 +44,6 @@ def read(filename):
 # numpy.loadtxt is super slow
 # Code adapted from <https://stackoverflow.com/a/8964779/353337>.
 def iter_loadtxt(infile, skiprows=0, comments=["#"], dtype=float, usecols=None):
-    pattern = re.compile("[a-zA-Z]*")
-
     def iter_func():
         for _ in range(skiprows):
             next(infile)
@@ -54,8 +51,8 @@ def iter_loadtxt(infile, skiprows=0, comments=["#"], dtype=float, usecols=None):
             line = line.decode("utf-8").strip()
             if line.startswith(comments):
                 continue
-            # remove all letters
-            items = pattern.sub("", line).strip().split()
+            # remove all text
+            items = line.split()[-3:]
             usecols_ = range(len(items)) if usecols is None else usecols
             for idx in usecols_:
                 yield dtype(items[idx])
@@ -161,10 +158,11 @@ def write(filename, mesh, binary=False):
             )
         )
     if len(mesh.cells) > 1:
-        keys = set(mesh.cells.keys())
-        keys.remove("triangle")
+        invalid = set(block.type for block in mesh.cells if block.type != "triangle")
         logging.warning(
-            "STL can only write triangle cells. Discarding {}.".format(", ".join(keys))
+            "STL can only write triangle cells. Discarding {}.".format(
+                ", ".join(invalid)
+            )
         )
 
     if mesh.points.shape[1] == 2:
