@@ -3,7 +3,7 @@ I/O for the TetGen file format, c.f.
 <https://wias-berlin.de/software/tetgen/fformats.node.html>
 """
 import logging
-import os
+import pathlib
 
 import numpy
 
@@ -14,18 +14,19 @@ from .._mesh import CellBlock, Mesh
 
 
 def read(filename):
-    base, ext = os.path.splitext(filename)
-    if ext == ".node":
+    filename = pathlib.Path(filename)
+    if filename.suffix == ".node":
         node_filename = filename
-        ele_filename = base + ".ele"
-    elif ext == ".ele":
-        node_filename = base + ".node"
+        ele_filename = filename.parent / (filename.stem + ".ele")
+    elif filename.suffix == ".ele":
+        node_filename = filename.parent / (filename.stem + ".node")
         ele_filename = filename
     else:
         raise ReadError()
 
     # read nodes
-    with open(node_filename) as f:
+    # TODO remove as_posix
+    with open(node_filename.as_posix()) as f:
         line = f.readline().strip()
         while len(line) == 0 or line[0] == "#":
             line = f.readline().strip()
@@ -51,7 +52,7 @@ def read(filename):
         points = points[:, 1:4]
 
     # read elements
-    with open(ele_filename) as f:
+    with open(ele_filename.as_posix()) as f:
         line = f.readline().strip()
         while len(line) == 0 or line[0] == "#":
             line = f.readline().strip()
@@ -71,13 +72,13 @@ def read(filename):
     return Mesh(points, [CellBlock("tetra", cells)])
 
 
-def write(filename, mesh, float_fmt=".15e"):
-    base, ext = os.path.splitext(filename)
-    if ext == ".node":
+def write(filename, mesh, float_fmt=".16e"):
+    filename = pathlib.Path(filename)
+    if filename.suffix == ".node":
         node_filename = filename
-        ele_filename = base + ".ele"
-    elif ext == ".ele":
-        node_filename = base + ".node"
+        ele_filename = filename.parent / (filename.stem + ".ele")
+    elif filename.suffix == ".ele":
+        node_filename = filename.parent / (filename.stem + ".node")
         ele_filename = filename
     else:
         raise WriteError("Must specify .node or .ele file. Got {}.".format(filename))
@@ -86,7 +87,8 @@ def write(filename, mesh, float_fmt=".15e"):
         raise WriteError("Can only write 3D points")
 
     # write nodes
-    with open(node_filename, "w") as fh:
+    # TODO remove .as_posix when requiring Python 3.6
+    with open(node_filename.as_posix(), "w") as fh:
         fh.write("# This file was created by meshio v{}\n".format(__version__))
         fh.write("{} {} {} {}\n".format(mesh.points.shape[0], 3, 0, 0))
         fmt = "{} " + " ".join(3 * ["{:" + float_fmt + "}"]) + "\n"
@@ -104,7 +106,8 @@ def write(filename, mesh, float_fmt=".15e"):
         )
 
     # write cells
-    with open(ele_filename, "w") as fh:
+    # TODO remove .as_posix when requiring Python 3.6
+    with open(ele_filename.as_posix(), "w") as fh:
         fh.write("# This file was created by meshio v{}\n".format(__version__))
         for cell_type, data in filter(lambda c: c.type == "tetra", mesh.cells):
             fh.write("{} {} {}\n".format(data.shape[0], 4, 0))
