@@ -7,7 +7,12 @@ from functools import reduce
 import numpy
 
 from ..__about__ import __version__
-from .._common import meshio_to_vtk_type, vtk_to_meshio_type
+from .._common import (
+    _meshio_to_vtk_order,
+    _vtk_to_meshio_order,
+    meshio_to_vtk_type,
+    vtk_to_meshio_type,
+)
 from .._exceptions import ReadError, WriteError
 from .._files import open_file
 from .._helpers import register
@@ -55,8 +60,8 @@ vtk_type_to_numnodes = numpy.array(
 )
 
 
-# These are all VTK data types. One sometimes finds 'vtktypeint64', but
-# this is ill-formed.
+# These are all VTK data types.
+# One sometimes finds 'vtktypeint64', but this is ill-formed.
 vtk_to_numpy_dtype_name = {
     "bit": "bool",
     "unsigned_char": "uint8",
@@ -65,7 +70,7 @@ vtk_to_numpy_dtype_name = {
     "short": "int16",
     "unsigned_int": "uint32",
     "int": "int32",
-    "unsigned_long": "int64",
+    "unsigned_long": "uint64",
     "long": "int64",
     "float": "float32",
     "double": "float64",
@@ -520,30 +525,29 @@ def translate_cells(data, types, cell_data_raw):
     if has_polygon:
         numnodes = numpy.empty(len(types), dtype=int)
         # If some polygons are in the VTK file, loop over the cells
-        nbcells = len(types)
+        numcells = len(types)
         offsets = numpy.empty(len(types), dtype=int)
         offsets[0] = 0
-        for idx in range(nbcells - 1):
+        for idx in range(numcells - 1):
             numnodes[idx] = data[offsets[idx]]
             offsets[idx + 1] = offsets[idx] + numnodes[idx] + 1
 
-        idx = nbcells - 1
+        idx = numcells - 1
         numnodes[idx] = data[offsets[idx]]
         if not numpy.all(numnodes == data[offsets]):
             raise ReadError()
 
         # TODO: cell_data
-        for idx in range(nbcells):
-            nbedges = data[offsets[idx]]
+        for idx, vtk_cell_type in enumerate(types):
             start = offsets[idx] + 1
-            end = start + numnodes[idx]
-            cell = data[start:end]
-            if nbedges == vtk_type_to_numnodes[meshio_to_vtk_type["triangle"]]:
-                cell_type = "triangle"
-            elif nbedges == vtk_type_to_numnodes[meshio_to_vtk_type["quad"]]:
-                cell_type = "quad"
-            else:
-                cell_type = "polygon" + str(nbedges)
+            cell_idx = start + _vtk_to_meshio_order(
+                vtk_cell_type, numnodes[idx], offsets.dtype
+            )
+            cell = data[cell_idx]
+
+            cell_type = vtk_to_meshio_type[vtk_cell_type]
+            if cell_type == "polygon":
+                cell_type += str(data[offsets[idx]])
 
             if len(cells) > 0 and cells[-1].type == cell_type:
                 cells[-1].data.append(cell)
@@ -571,7 +575,8 @@ def translate_cells(data, types, cell_data_raw):
         for start, end in zip(b[:-1], b[1:]):
             meshio_type = vtk_to_meshio_type[types[start]]
             n = data[offsets[start]]
-            indices = numpy.add.outer(offsets[start:end], numpy.arange(1, n + 1))
+            cell_idx = 1 + _vtk_to_meshio_order(types[start], n, dtype=offsets.dtype)
+            indices = numpy.add.outer(offsets[start:end], cell_idx)
             cells.append(CellBlock(meshio_type, data[indices]))
             for name, d in cell_data_raw.items():
                 if name not in cell_data:
@@ -661,27 +666,35 @@ def _write_points(f, points, binary):
 
 def _write_cells(f, cells, binary):
     total_num_cells = sum([len(c.data) for c in cells])
-    total_num_idx = sum([numpy.prod(c.data.shape) for c in cells])
+    total_num_idx = sum([c.data.size for c in cells])
     # For each cell, the number of nodes is stored
     total_num_idx += total_num_cells
     f.write("CELLS {} {}\n".format(total_num_cells, total_num_idx).encode("utf-8"))
     if binary:
         for c in cells:
             n = c.data.shape[1]
+            cell_idx = _meshio_to_vtk_order(c.type, n)
             dtype = numpy.dtype(">i4")
             # One must force endianness here:
             # <https://github.com/numpy/numpy/issues/15088>
             numpy.column_stack(
-                [numpy.full(c.data.shape[0], n, dtype=dtype), c.data.astype(dtype)],
+                [
+                    numpy.full(c.data.shape[0], n, dtype=dtype),
+                    c.data[:, cell_idx].astype(dtype),
+                ],
             ).astype(dtype).tofile(f, sep="")
         f.write(b"\n")
     else:
         # ascii
         for c in cells:
             n = c.data.shape[1]
+            cell_idx = _meshio_to_vtk_order(c.type, n)
             # prepend a column with the value n
             numpy.column_stack(
-                [numpy.full(c.data.shape[0], n, dtype=c.data.dtype), c.data]
+                [
+                    numpy.full(c.data.shape[0], n, dtype=c.data.dtype),
+                    c.data[:, cell_idx],
+                ]
             ).tofile(f, sep="\n")
             f.write(b"\n")
 
