@@ -6,6 +6,7 @@ I/O for VTU.
 import base64
 import logging
 import lzma
+import re
 import sys
 import zlib
 
@@ -13,6 +14,8 @@ import numpy
 
 from ..__about__ import __version__
 from .._common import (
+    _meshio_to_vtk_order,
+    _vtk_to_meshio_order,
     meshio_to_vtk_type,
     num_nodes_per_cell,
     raw_from_cell_data,
@@ -47,15 +50,49 @@ def _cells_from_data(connectivity, offsets, types, cell_data_raw):
             meshio_type = vtk_to_meshio_type[types[start]]
         except KeyError:
             raise ReadError("File contains cells that meshio cannot handle.")
-        n = num_nodes_per_cell[meshio_type]
-        indices = numpy.add.outer(
-            offsets[start:end], numpy.arange(-n, 0, dtype=offsets.dtype)
-        )
-        cells.append(CellBlock(meshio_type, connectivity[indices]))
-        for name, d in cell_data_raw.items():
-            if name not in cell_data:
-                cell_data[name] = []
-            cell_data[name].append(d[start:end])
+        if meshio_type == "polygon" or meshio_type == "polyhedron":
+            # Polygons and polyhedra have unknown and varying number of nodes per cell.
+
+            # Index where the previous block of cells stopped. Needed to know the number
+            # of nodes for the first cell in the block.
+            if start == 0:
+                # This is the very start of the offset array
+                first_node = 0
+            else:
+                # First node is the end of the offset for the previous block
+                first_node = offsets[start - 1]
+
+            # Start of the cell-node relation for each cell in this block
+            start_cn = numpy.hstack((first_node, offsets[start:end]))
+            # Find the size of each cell
+            size = numpy.diff(start_cn)
+
+            # Loop over all cell sizes, find all cells with this size, and assign
+            # connectivity
+            for sz in numpy.unique(size):
+                items = numpy.where(size == sz)[0]
+                indices = numpy.add.outer(
+                    start_cn[items + 1],
+                    _vtk_to_meshio_order(types[start], sz, dtype=offsets.dtype) - sz,
+                )
+                cells.append(CellBlock(meshio_type + str(sz), connectivity[indices]))
+                # Store cell data for this set of cells
+                for name, d in cell_data_raw.items():
+                    if name not in cell_data:
+                        cell_data[name] = []
+                    cell_data[name].append(d[start + items])
+        else:
+            # Same number of nodes per cell
+            n = num_nodes_per_cell[meshio_type]
+            indices = numpy.add.outer(
+                offsets[start:end],
+                _vtk_to_meshio_order(types[start], n, dtype=offsets.dtype) - n,
+            )
+            cells.append(CellBlock(meshio_type, connectivity[indices]))
+            for name, d in cell_data_raw.items():
+                if name not in cell_data:
+                    cell_data[name] = []
+                cell_data[name].append(d[start:end])
 
     return cells, cell_data
 
@@ -67,7 +104,10 @@ def _organize_cells(point_offsets, cells, cell_data_raw):
     out_cells = []
     for offset, cls, cdr in zip(point_offsets, cells, cell_data_raw):
         cls, cell_data = _cells_from_data(
-            cls["connectivity"], cls["offsets"], cls["types"], cdr
+            cls["connectivity"].ravel(),
+            cls["offsets"].ravel(),
+            cls["types"].ravel(),
+            cdr,
         )
         for c in cls:
             out_cells.append(CellBlock(c.type, c.data + offset))
@@ -101,6 +141,84 @@ def get_grid(root):
     return grid, appended_data
 
 
+def _parse_raw_binary(filename):
+    from xml.etree import ElementTree as ET
+
+    with open(filename, "rb") as f:
+        raw = f.read()
+
+    try:
+        i_start = re.search(re.compile(b'<AppendedData[^>]+(?:">)'), raw).end()
+        i_stop = raw.find(b"</AppendedData>")
+    except Exception:
+        raise ReadError()
+
+    header = raw[:i_start].decode()
+    footer = raw[i_stop:].decode()
+    data = raw[i_start:i_stop].split(b"_", 1)[1].rsplit(b"\n", 1)[0]
+
+    root = ET.fromstring(header + footer)
+
+    dtype = vtu_to_numpy_type[root.get("header_type", "UInt32")]
+    if "byte_order" in root.attrib:
+        dtype = dtype.newbyteorder(
+            "<" if root.get("byte_order") == "LittleEndian" else ">"
+        )
+
+    appended_data_tag = root.find("AppendedData")
+    appended_data_tag.set("encoding", "base64")
+
+    if "compressor" in root.attrib:
+        c = {"vtkLZMADataCompressor": lzma, "vtkZLibDataCompressor": zlib}[
+            root.get("compressor")
+        ]
+        root.attrib.pop("compressor")
+
+        # raise ReadError("Compressed raw binary VTU files not supported.")
+        arrays = ""
+        i = 0
+        while i < len(data):
+            da_tag = root.find(".//DataArray[@offset='%d']" % i)
+            da_tag.set("offset", "%d" % len(arrays))
+
+            num_blocks = int(numpy.frombuffer(data[i : i + dtype.itemsize], dtype)[0])
+            num_header_items = 3 + num_blocks
+            num_header_bytes = num_header_items * dtype.itemsize
+            header = numpy.frombuffer(data[i : i + num_header_bytes], dtype)
+
+            block_data = b""
+            j = 0
+            for k in range(num_blocks):
+                block_size = int(header[k + 3])
+                block_data += c.decompress(
+                    data[
+                        i + j + num_header_bytes : i + j + block_size + num_header_bytes
+                    ]
+                )
+                j += block_size
+
+            block_size = numpy.array([len(block_data)]).astype(dtype).tobytes()
+            arrays += base64.b64encode(block_size + block_data).decode()
+
+            i += j + num_header_bytes
+
+    else:
+        arrays = ""
+        i = 0
+        while i < len(data):
+            da_tag = root.find(".//DataArray[@offset='%d']" % i)
+            da_tag.set("offset", "%d" % len(arrays))
+
+            block_size = int(numpy.frombuffer(data[i : i + dtype.itemsize], dtype)[0])
+            arrays += base64.b64encode(
+                data[i : i + block_size + dtype.itemsize]
+            ).decode()
+            i += block_size + dtype.itemsize
+
+    appended_data_tag.text = "_" + arrays
+    return root
+
+
 vtu_to_numpy_type = {
     "Float32": numpy.dtype(numpy.float32),
     "Float64": numpy.dtype(numpy.float64),
@@ -123,11 +241,14 @@ class VtuReader:
     """
 
     def __init__(self, filename):  # noqa: C901
-        import xml.etree.ElementTree as ET
+        from xml.etree import ElementTree as ET
 
         parser = ET.XMLParser()
-        tree = ET.parse(filename, parser)
-        root = tree.getroot()
+        try:
+            tree = ET.parse(str(filename), parser)
+            root = tree.getroot()
+        except ET.ParseError:
+            root = _parse_raw_binary(str(filename))
 
         if root.tag != "VTKFile":
             raise ReadError()
@@ -138,6 +259,10 @@ class VtuReader:
                 "Unknown VTU file version '{}'.".format(root.attrib["version"])
             )
 
+        # fix empty NumberOfComponents attributes as produced by Firedrake
+        for da_tag in root.findall(".//DataArray[@NumberOfComponents='']"):
+            da_tag.attrib.pop("NumberOfComponents")
+
         if "compressor" in root.attrib:
             assert root.attrib["compressor"] in [
                 "vtkLZMADataCompressor",
@@ -242,9 +367,7 @@ class VtuReader:
         if len(cell_data_raw) != len(cells):
             raise ReadError()
 
-        point_offsets = (
-            numpy.cumsum([pts.shape[0] for pts in points]) - points[0].shape[0]
-        )
+        point_offsets = numpy.cumsum([0] + [pts.shape[0] for pts in points][:-1])
 
         # Now merge across pieces
         if not points:
@@ -273,7 +396,17 @@ class VtuReader:
                 "<" if self.byte_order == "LittleEndian" else ">"
             )
         num_bytes_per_item = numpy.dtype(dtype).itemsize
-        # total_num_bytes = numpy.frombuffer(byte_string[:num_bytes_per_item], dtype)[0]
+        total_num_bytes = int(
+            numpy.frombuffer(byte_string[:num_bytes_per_item], dtype)[0]
+        )
+
+        # Check if block size was decoded separately
+        # (so decoding stopped after block size due to padding)
+        if len(byte_string) == num_bytes_per_item:
+            header_len = len(base64.b64encode(byte_string))
+            byte_string = base64.b64decode(data[header_len:])
+        else:
+            byte_string = byte_string[num_bytes_per_item:]
 
         # Read the block data; multiple blocks possible here?
         dtype = vtu_to_numpy_type[data_type]
@@ -281,7 +414,7 @@ class VtuReader:
             dtype = dtype.newbyteorder(
                 "<" if self.byte_order == "LittleEndian" else ">"
             )
-        return numpy.frombuffer(byte_string[num_bytes_per_item:], dtype=dtype)
+        return numpy.frombuffer(byte_string[:total_num_bytes], dtype=dtype)
 
     def read_compressed_binary(self, data, data_type):
         # first read the block size; it determines the size of the header
@@ -428,19 +561,23 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
     # swap the data to match the system byteorder
     # Don't use byteswap to make sure that the dtype is changed; see
     # <https://github.com/numpy/numpy/issues/10372>.
-    points = mesh.points.astype(mesh.points.dtype.newbyteorder("="))
+    points = mesh.points.astype(mesh.points.dtype.newbyteorder("="), copy=False)
+    for k, (cell_type, data) in enumerate(mesh.cells):
+        mesh.cells[k] = CellBlock(
+            cell_type, data.astype(data.dtype.newbyteorder("="), copy=False)
+        )
     for key, data in mesh.point_data.items():
-        mesh.point_data[key] = data.astype(data.dtype.newbyteorder("="))
+        mesh.point_data[key] = data.astype(data.dtype.newbyteorder("="), copy=False)
     for data in mesh.cell_data.values():
         for k, dat in enumerate(data):
-            data[k] = dat.astype(dat.dtype.newbyteorder("="))
+            data[k] = dat.astype(dat.dtype.newbyteorder("="), copy=False)
     for key, data in mesh.field_data.items():
-        mesh.field_data[key] = data.astype(data.dtype.newbyteorder("="))
+        mesh.field_data[key] = data.astype(data.dtype.newbyteorder("="), copy=False)
 
-    def numpy_to_xml_array(parent, name, fmt, data):
-        da = ET.SubElement(
-            parent, "DataArray", type=numpy_to_vtu_type[data.dtype], Name=name
-        )
+    def numpy_to_xml_array(parent, name, data):
+        vtu_type = numpy_to_vtu_type[data.dtype]
+        fmt = "{:.11e}" if vtu_type.startswith("Float") else "{:d}"
+        da = ET.SubElement(parent, "DataArray", type=vtu_type, Name=name)
         if len(data.shape) == 2:
             da.set("NumberOfComponents", "{}".format(data.shape[1]))
         if binary:
@@ -449,7 +586,7 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
                 # compressed write
                 def text_writer(f):
                     max_block_size = 32768
-                    data_bytes = data.tostring()
+                    data_bytes = data.tobytes()
 
                     # round up
                     num_blocks = -int(-len(data_bytes) // max_block_size)
@@ -474,18 +611,18 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
                         + [len(b) for b in compressed_blocks],
                         dtype=vtu_to_numpy_type[header_type],
                     )
-                    f.write(base64.b64encode(header.tostring()).decode())
+                    f.write(base64.b64encode(header.tobytes()).decode())
                     f.write(base64.b64encode(b"".join(compressed_blocks)).decode())
 
             else:
                 # uncompressed write
                 def text_writer(f):
-                    data_bytes = data.tostring()
+                    data_bytes = data.tobytes()
                     # collect header
                     header = numpy.array(
                         len(data_bytes), dtype=vtu_to_numpy_type[header_type]
                     )
-                    f.write(base64.b64encode(header.tostring() + data_bytes).decode())
+                    f.write(base64.b64encode(header.tobytes() + data_bytes).decode())
 
         else:
             da.set("format", "ascii")
@@ -519,13 +656,18 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
     # points
     if points is not None:
         pts = ET.SubElement(piece, "Points")
-        numpy_to_xml_array(pts, "Points", "{:.11e}", points)
+        numpy_to_xml_array(pts, "Points", points)
 
     if mesh.cells is not None:
         cls = ET.SubElement(piece, "Cells")
 
         # create connectivity, offset, type arrays
-        connectivity = numpy.concatenate([v.data.reshape(-1) for v in mesh.cells])
+        connectivity = numpy.concatenate(
+            [
+                v.data[:, _meshio_to_vtk_order(v.type, v.data.shape[1])].reshape(-1)
+                for v in mesh.cells
+            ]
+        )
 
         # offset (points to the first element of the next cell)
         offsets = [
@@ -538,23 +680,37 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
         offsets = numpy.concatenate(offsets)
 
         # types
+        types_array = []
+        for k, v in mesh.cells:
+            # For polygon and polyhedron grids, the number of nodes is part of the cell
+            # type key. This part must be stripped away.
+            if k[:7] == "polygon":
+                key_ = k[:7]
+            elif k[:10] == "polyhedron":
+                key_ = k[:10]
+            else:
+                # No special treatment
+                key_ = k
+            types_array.append(numpy.full(len(v), meshio_to_vtk_type[key_]))
+
         types = numpy.concatenate(
-            [numpy.full(len(v), meshio_to_vtk_type[k]) for k, v in mesh.cells]
+            types_array
+            # [numpy.full(len(v), meshio_to_vtk_type[k]) for k, v in mesh.cells]
         )
 
-        numpy_to_xml_array(cls, "connectivity", "{:d}", connectivity)
-        numpy_to_xml_array(cls, "offsets", "{:d}", offsets)
-        numpy_to_xml_array(cls, "types", "{:d}", types)
+        numpy_to_xml_array(cls, "connectivity", connectivity)
+        numpy_to_xml_array(cls, "offsets", offsets)
+        numpy_to_xml_array(cls, "types", types)
 
     if mesh.point_data:
         pd = ET.SubElement(piece, "PointData")
         for name, data in mesh.point_data.items():
-            numpy_to_xml_array(pd, name, "{:.11e}", data)
+            numpy_to_xml_array(pd, name, data)
 
     if mesh.cell_data:
         cd = ET.SubElement(piece, "CellData")
         for name, data in raw_from_cell_data(mesh.cell_data).items():
-            numpy_to_xml_array(cd, name, "{:.11e}", data)
+            numpy_to_xml_array(cd, name, data)
 
     # write_xml(filename, vtk_file, pretty_xml)
     tree = ET.ElementTree(vtk_file)
