@@ -3,12 +3,12 @@ I/O for XDMF.
 http://www.xdmf.org/index.php/XDMF_Model_and_Format
 """
 import os
-import xml.etree.ElementTree as ET
 from io import BytesIO
+from xml.etree import ElementTree as ET
 
 import numpy
 
-from .._common import CDATA, cell_data_from_raw, raw_from_cell_data, write_xml
+from .._common import cell_data_from_raw, raw_from_cell_data, write_xml
 from .._exceptions import ReadError, WriteError
 from .._helpers import register
 from .._mesh import CellBlock, Mesh
@@ -51,6 +51,8 @@ class XdmfReader:
         return self.read_xdmf3(root)
 
     def _read_data_item(self, data_item, root=None):
+        import h5py
+
         reference = data_item.get("Reference")
         if reference:
             xpath = (data_item.text if reference == "XML" else reference).strip()
@@ -141,7 +143,7 @@ class XdmfReader:
             raise ReadError()
 
         points = None
-        cells = {}
+        cells = []
         point_data = {}
         cell_data_raw = {}
         field_data = {}
@@ -162,8 +164,8 @@ class XdmfReader:
                         )
                     )
                 else:
-                    meshio_type = xdmf_to_meshio_type[topology_type]
-                    cells[meshio_type] = self._read_data_item(data_items[0])
+                    data = self._read_data_item(data_items[0])
+                    cells.append(CellBlock(xdmf_to_meshio_type[topology_type], data))
 
             elif c.tag == "Geometry":
                 if c.get("GeometryType") not in (None, "XYZ"):
@@ -315,6 +317,8 @@ class XdmfWriter:
     def __init__(
         self, filename, mesh, data_format="HDF", compression="gzip", compression_opts=4
     ):
+        import h5py
+
         if data_format not in ["XML", "Binary", "HDF"]:
             raise WriteError(
                 "Unknown XDMF data format "
@@ -335,12 +339,12 @@ class XdmfWriter:
 
         domain = ET.SubElement(xdmf_file, "Domain")
         grid = ET.SubElement(domain, "Grid", Name="Grid")
-        information = ET.SubElement(
-            grid, "Information", Name="Information", Value=str(len(mesh.field_data))
-        )
+        # information = ET.SubElement(
+        #     grid, "Information", Name="Information", Value=str(len(mesh.field_data))
+        # )
 
         self.points(grid, mesh.points)
-        self.field_data(mesh.field_data, information)
+        # self.field_data(mesh.field_data, information)
         self.cells(mesh.cells, grid)
         self.point_data(mesh.point_data, grid)
         self.cell_data(mesh.cell_data, grid)
@@ -410,6 +414,7 @@ class XdmfWriter:
                 "Topology",
                 TopologyType=xdmf_type,
                 NumberOfElements=str(num_cells),
+                NodesPerElement=str(cells[0].data.shape[1]),
             )
             dt, prec = numpy_to_xdmf_dtype[cells[0].data.dtype.name]
             dim = "{} {}".format(*cells[0].data.shape)
@@ -505,26 +510,25 @@ class XdmfWriter:
             )
             data_item.text = self.numpy_to_xml_string(data)
 
-    def field_data(self, field_data, information):
-        info = ET.Element("main")
-        for name, data in field_data.items():
-            data_item = ET.SubElement(info, "map", key=name, dim=str(data[1]))
-            data_item.text = str(data[0])
-        # information.text = ET.CDATA(ET.tostring(info))
-        information.append(CDATA(ET.tostring(info).decode("utf-8")))
+    # The original idea was to implement field data as XML CDATA. Unfortunately, in
+    # Python's XML, CDATA handled poorly. There are workarounds, e.g.,
+    # <https://stackoverflow.com/a/30019607/353337>, but those mess around with
+    # ET._serialize_xml and lead to bugs elsewhere
+    # <https://github.com/nschloe/meshio/issues/806>.
+    # def field_data(self, field_data, information):
+    #     info = ET.Element("main")
+    #     for name, data in field_data.items():
+    #         data_item = ET.SubElement(info, "map", key=name, dim=str(data[1]))
+    #         data_item.text = str(data[0])
+    #     information.text = ET.CDATA(ET.tostring(info))
+    #     information.append(CDATA(ET.tostring(info).decode("utf-8")))
 
 
 def write(*args, **kwargs):
     XdmfWriter(*args, **kwargs)
 
 
-try:
-    import h5py
-# Use ModuleNotFoundError when dropping support for Python 3.5
-except ImportError:
-    pass
-else:
-    # TODO register all xdmf except hdf outside this try block
-    register(
-        "xdmf", [".xdmf", ".xmf"], read, {"xdmf": write},
-    )
+# TODO register all xdmf except hdf outside this try block
+register(
+    "xdmf", [".xdmf", ".xmf"], read, {"xdmf": write},
+)
