@@ -1,7 +1,7 @@
 import os
 import warnings
-import xml.etree.ElementTree as ET
 from io import BytesIO
+from xml.etree import ElementTree as ET
 
 import numpy
 
@@ -90,7 +90,7 @@ class TimeSeriesReader:
         grid = self.mesh_grid
 
         points = None
-        cells = {}
+        cells = []
 
         for c in grid:
             if c.tag == "Topology":
@@ -113,8 +113,7 @@ class TimeSeriesReader:
                 if cell_type == "Mixed":
                     cells = translate_mixed_cells(data)
                 else:
-                    meshio_type = xdmf_to_meshio_type[cell_type]
-                    cells[meshio_type] = data
+                    cells.append(CellBlock(xdmf_to_meshio_type[cell_type], data))
 
             elif c.tag == "Geometry":
                 try:
@@ -288,6 +287,7 @@ class TimeSeriesWriter:
         write_xml(self.filename, self.xdmf_file)
 
     def write_data(self, t, point_data=None, cell_data=None):
+        cell_data = {} if cell_data is None else cell_data
         # <Grid>
         #   <xi:include xpointer="xpointer(//Grid[@Name=&quot;TimeSeries_phi&quot;]/Grid[1]/*[self::Topology or self::Geometry])" />
         #   <Time Value="3.3333333333333335e-05" />
@@ -392,20 +392,21 @@ class TimeSeriesWriter:
             )
             data_item.text = self.numpy_to_xml_string(cells[0].data)
         elif len(cells) > 1:
-            total_num_cells = sum(c.shape[0] for c in cells.values())
+            total_num_cells = sum(c.data.shape[0] for c in cells)
             topo = ET.SubElement(
                 grid,
                 "Topology",
                 TopologyType="Mixed",
                 NumberOfElements=str(total_num_cells),
             )
-            total_num_cell_items = sum(numpy.prod(c.shape) for c in cells.values())
+            total_num_cell_items = sum(numpy.prod(c.data.shape) for c in cells)
             dim = total_num_cell_items + total_num_cells
             # Lines translate to Polylines, and one needs to specify the exact
             # number of nodes. Hence, prepend 2.
-            if "line" in cells:
-                cells["line"] = numpy.insert(cells["line"], 0, 2, axis=1)
-                dim += len(cells["line"])
+            for c in cells:
+                if c.type == "line":
+                    c.data[:] = numpy.insert(c.data, 0, 2, axis=1)
+                    dim += len(c.data)
             dim = str(dim)
             cd = numpy.concatenate(
                 [
@@ -413,7 +414,7 @@ class TimeSeriesWriter:
                     numpy.insert(
                         value, 0, meshio_type_to_xdmf_index[key], axis=1
                     ).flatten()
-                    for key, value in cells.items()
+                    for key, value in cells
                 ]
             )
             dt, prec = numpy_to_xdmf_dtype[cd.dtype.name]
