@@ -1,9 +1,9 @@
-import os
+import pathlib
 
+import helpers
 import numpy
 import pytest
 
-import helpers
 import meshio
 
 
@@ -24,8 +24,9 @@ def test_obj(mesh):
     "filename, ref_sum, ref_num_cells", [("elephav.obj", 3.678372172450000e05, 1148)]
 )
 def test_reference_file(filename, ref_sum, ref_num_cells):
-    this_dir = os.path.dirname(os.path.abspath(__file__))
-    filename = os.path.join(this_dir, "meshes", "obj", filename)
+    this_dir = pathlib.Path(__file__).resolve().parent
+    filename = this_dir / "meshes" / "obj" / filename
+
     mesh = meshio.read(filename)
     tol = 1.0e-5
     s = numpy.sum(mesh.points)
